############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import time, warnings, types, weakref

import ZSI.TC
from ZSI import  _seqtypes
from ZSI.TC import _get_global_element_declaration as GED
from ZSI.TC import ElementDeclaration, ComplexType, SimpleType

try:
    from ZSI.schema import WrapImmutable
except: 
    # TODO: reorg, remove when ZSI-2.0.0 final released
    from ZSI.TC import Wrap as WrapImmutable

from ZSI.writer import SoapWriter
from ZSI.parse import ParsedSoap
from ZSI.wstools.Namespaces import OASIS, DSIG
from ZSI.wstools.Utility import SplitQName

from twisted.python import log
from twisted.internet import reactor
from ZSI.wstools.Utility import SplitQName

from pyGridWare.wsrf.faults.PropertiesFaults import *
from pyGridWare.utility.dom import DomletteElementProxy as EProxy

# Set to one year.
DEFAULT_TERMINATION = 365*24*60*60

'''
Classes:
    NoResourcePropertyError
    DeclarationError
    ResourceProperties
    NotificationProducerRP
    ScheduledResourceTerminationRP
    SubscriptionManagerRP
'''


class NoResourcePropertyError(AttributeError):
    '''No Such Resource Property exists in ResourceProperty declararion.
    '''
class DeclarationError(NotImplementedError):
    '''delcaration not declared.
    '''

class ResourceProperties(object):
    '''ResourceProperties must be a GED, and it contains any number 
    references to other GEDs. This class must be subclassed, and
    declaration must be specified.

    class variables
        declaration -- must be a class type representing a GED.
        
    instance variables
        topicListeners -- list of callbacks.  whenever a resource 
            property is "set" all items are called.
    '''
    declaration = None

    def __init__(self):
        self.CheckTypecode()
        self.topicListeners = []
        self._typecode = self.declaration
        self._pyobj = self._typecode.pyclass()

    def __str__(self):
        sw = SoapWriter(header=False, envelope=False)
        sw.serialize(self._pyobj, self._typecode)
        return str(sw)

    def setUp(self, xml=None):
        '''xml -- string representation of RP XML instance.
        '''
        if xml is None: return
        ps = ParsedSoap(xml, envelope=False)
        self._pyobj = ps.Parse(self._typecode)

    def getContext(self, context):
        return self._context()
    def setContext(self, context):
        self._context = weakref.ref(context)
    context = property(getContext, setContext, None, 
        'Set properties resource context')
    
    def query(self, exp, response):
        '''Query ResourceProperties with QueryExpression.
        exp -- QueryExpression element.
        response -- QueryResourcePropertiesResponse
        '''
        if exp._attrs['Dialect'] != DSIG.XPATH:
            raise UnknownQueryExpressionDialectFault(\
                timestamp=time.gmtime(),
                originator=None,
                errorcode=None,
                description=None,
                faultcause=('Unknown Dialect: %s' %exp._attrs['Dialect'],),
            )

        sw = SoapWriter(header=False, envelope=False, outputclass=EProxy)
        try:
            sw.serialize(self._pyobj, self._pyobj.typecode)
        except Exception, ex:
            raise QueryEvaluationErrorFault(\
                timestamp=time.gmtime(), 
                originator=None,
                errorcode=None,
                description=None,
                faultcause=ex.args,
            )

        try:
            result = sw.dom.evaluate(exp._text)
        except Exception, ex:
            raise InvalidQueryExpressionFault(\
                timestamp=time.gmtime(), 
                originator=None,
                errorcode=None,
                description=None,
                faultcause=('Dialect: %s' %exp._attrs['Dialect'],),
            )
            
        # TODO:
        # Must provide at least one any "element" because minOccurs=1
        # All examples use text content w/o using <any>.  Must load
        # pyGridWare.patch
        response._text = '%s' %result

    def getResourcePropertiesDocument(self):
        '''ResourceProperties must be a global element declaration.
        '''
        return self._pyobj

    def getResourceProperty(self, namespaceURI, localName):
        '''Resource must be a global element declaration, 
        and so assume that it is either the ResourceProperty
        element itself, or one of the element references in
        its model group.
        '''
        tc = self._pyobj.typecode
        log.msg("Resource Property Definition(%s, %s): " %
                    (tc.nspname,tc.pname), debug=True)
        
        if tc.nspname == namespaceURI and  tc.pname == localName:
            log.msg("Grab Resource Property Definition(%s, %s): ", debug=True)
            return self._pyobj
            
        what = element = None
        f = lambda w: w.nspname==namespaceURI and w.pname==localName
        ofwhat = filter(f, getattr(tc, 'ofwhat', tuple()))
        if len(ofwhat) == 1:
            what = ofwhat[0]
            # Check for Mirage
            if callable(what): what = what()
            element = getattr(self._pyobj, what.aname, None)
            
        if element is None:
            # could mean duplicate typecode entries, or 
            raise InvalidResourcePropertyQNameFault(timestamp=time.gmtime(),
                    faultcause=(
                        'found %d element(s) (%s,%s) in resourceProperty(%s,%s) '%
                        (len(ofwhat), namespaceURI,localName,tc.nspname,tc.pname),),)
        
        log.msg("Resource Property Reference(%s, %s): " %
                    (what.nspname,what.pname), debug=True)
        
        # make sure element is self-describing
        if what.maxOccurs == 1:
            element = [element]
        
        pyobjs = []
        append = pyobjs.append
        for e in element:
            if hasattr(e, 'typecode'):
                if e.typecode is None: 
                    e.typecode = what
                append(e)
            elif isinstance(what, SimpleType):
                pyobj = WrapImmutable(e, what)
                append(pyobj)
            else:
                e.typecode = what
                append(e)
        
        if what.maxOccurs == 1:
            return pyobjs[0]
            
        return pyobjs

    def setResourceProperty(self, namespaceURI, localName, pyobj):
        what = self._typecode
        for what in what.ofwhat:
            if what.nspname == namespaceURI and  what.pname == localName:
                break
        else:
            what = self._typecode
            raise SetResourcePropertyRequestFailedFault(timestamp=time.gmtime(),
                faultcause=('no element (%s,%s) in resourceProperty(%s,%s) '\
                %(namespaceURI,localName,what.nspname,what.pname),),
            )

        setattr(self._pyobj, what.aname, pyobj)
        if pyobj is None: 
            log.msg('pyobj is None', debug=True)
            if what.minOccurs == 0 or what.nillable:
                if len(self.topicListeners) > 0:
                    warnings.warn(\
                        'Not notifying TopicListeners of change of property(%s,%s) to None'\
                        %(what.nspname,what.pname)
                    )
                return
            warnings.warn('missing instance of class %s minOccurs=%d, nillable=%d' \
                %(what.__class__, what.minOccurs, what.nillable)
            )
            raise SetResourcePropertyRequestFailedFault(timestamp=time.gmtime(),
                faultcause=('missing instance of element(%s, %s)' %(what.nspname, what.pname)),
            )


        #
        # Need to time out on callbacks.
        #
        pyobj = self.getResourceProperty(namespaceURI, localName)
        for cb in self.topicListeners:
            log.msg('Execute Callback -- %s' %str(cb), debug=True)
            reactor.callLater(0.0001, cb, pyobj, namespaceURI, localName)

    def CheckTypecode(cls, typecode=None):
        '''Recursive base check to make sure declaration is a superset of all declarations 
        of RP subclasses.  If typecode is None, grab the class variable typecode and
        check it against its base classes. 
        '''
        ctypecode = cls.GetTypecode()
        if ctypecode is None: 
            return

        for base in filter(lambda b: issubclass(b, ResourceProperties), cls.__bases__):
            base.CheckTypecode(typecode or ctypecode)

        if typecode is None: 
            return

        ofwhat = map(lambda w: (w.nspname,w.pname), typecode.ofwhat)
        for what in ctypecode.ofwhat:
            if (what.nspname,what.pname) not in ofwhat:
                raise ResourceUnknownFault(timestamp=time.gmtime(),
                    faultcause=('RP declaration must declare (%s,%s)' %(what.nspname,what.pname),),
                )
    CheckTypecode = classmethod(CheckTypecode)

    def GetTypecode(cls):
        typecode = cls.declaration
        if typecode is not None and not isinstance(typecode, ElementDeclaration):
             raise DeclarationError, \
                 'ResourceProperties declaration variable must be instance of ElementDeclaration'\

        return typecode
    GetTypecode = classmethod(GetTypecode)


class NotificationProducerRP(ResourceProperties):
    '''NotificationProducer Resource Properties.
    Topic 
    FixedTopicSet 
    TopicExpressionDialects 
    '''
    declaration = GED(OASIS.BASENOTIFICATION, 'NotificationProducerRP')

    def __init__(self):
        '''SubscriptionManager Singleton is the NotificationProducer, it will 
        set the Topic, FixedTopicSet, and TopicExpressionDialects RPs.
        '''
        from pyGridWare.wsrf.notification.SubscriptionManager import SubscriptionManager
        assert SubscriptionManager._isInstantiated(),\
            "SubscriptionManager is not instantiated"
        manager = SubscriptionManager.getInstance()
        manager.initialize(self)

    def getTopic(self):
        return self.getResourceProperty(OASIS.BASENOTIFICATION, "Topic")
    def setTopic(self, value):
        self.setResourceProperty(OASIS.BASENOTIFICATION, "Topic", value)
    Topic = property(getTopic, setTopic, None, 'Topic')

    def getFixedTopicSet(self):
        return self.getResourceProperty(OASIS.BASENOTIFICATION, "FixedTopicSet")
    def setFixedTopicSet(self, value):
        self.setResourceProperty(OASIS.BASENOTIFICATION, "FixedTopicSet", value)
    FixedTopicSet = property(getFixedTopicSet, setFixedTopicSet, None, 'Fixed Topic Set')

    def getTopicExpressionDialects(self):
        return self.getResourceProperty(OASIS.BASENOTIFICATION, "TopicExpressionDialects")
    def setTopicExpressionDialects(self, value):
        self.setResourceProperty(OASIS.BASENOTIFICATION, "TopicExpressionDialects", value)
    TopicExpressionDialects = property(getTopicExpressionDialects, setTopicExpressionDialects, None, \
        'Topic Expression Dialects')

    def addTopicListener(self, method, sub, stub):
        ''' Need to hold onto DOM to make this really work
        specification is retarded, uses text content to pass
        in a qualifiedName.

        sub -- Subscription 
        stub -- NotificationProducer stub
        '''
        assert callable(method), 'Callback must be callable.'
        topic = sub.topic
        if not getattr(topic, '_text', None): return
        prefix,name = SplitQName(topic._text)
        log.msg("TOPIC: %s:%s" %(prefix,name), debug=True)

        def cb(value, nspname, pname):
            if pname != name: return
            return method(value, sub, stub)

        self.topicListeners.append(cb)



class ScheduledResourceTerminationRP(ResourceProperties):
    '''
                    self._CurrentTime = None
                    self._TerminationTime = None
    '''
    declaration = GED(OASIS.BASENOTIFICATION, 'ScheduledResourceTerminationRP')

    def __init__(self):
        self.currentTime = time.gmtime()
        self.terminationTime = time.gmtime(time.time()+DEFAULT_TERMINATION)

    def getCurrentTime(self):
        return self.getResourceProperty(OASIS.LIFETIME, "CurrentTime")
    def setCurrentTime(self, value):
        assert isinstance(value, tuple) or isinstance(value, time.struct_time) and len(value) == 9,\
            'expecting 9 item tuple got len(%d), %s' %(len(value),str(value))
        self.setResourceProperty(OASIS.LIFETIME, "CurrentTime", tuple(value))
    currentTime = property(getCurrentTime, setCurrentTime, None, 'current time')

    def getTerminationTime(self):
        return self.getResourceProperty(OASIS.LIFETIME, "TerminationTime")
    def setTerminationTime(self, value):
        assert isinstance(value, tuple) or isinstance(value, time.struct_time) and len(value) == 9,\
            'expecting 9 item tuple got len(%d), %s' %(len(value),str(value))
        self.setResourceProperty(OASIS.LIFETIME, "TerminationTime", tuple(value))
    terminationTime = property(getTerminationTime, setTerminationTime, None, 'Termination time')



class SubscriptionManagerRP(ScheduledResourceTerminationRP):
    '''
    CurrentTime 
    TerminationTime 
    ConsumerReference 
    TopicExpression 
    UseNotify 
    Precondition 
    Selector 
    SubscriptionPolicy 
    CreationTime 
    '''
    declaration = GED(OASIS.BASENOTIFICATION, 'SubscriptionManagerRP')

