##########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import ez_setup
ez_setup.use_setuptools()

import os, sys, warnings, pydoc, inspect
from distutils import log
from os.path import isdir, isfile, join
from os.path import sep as SEP
from ConfigParser import ConfigParser, NoOptionError
from setuptools import setup, find_packages
from setuptools.dist import Distribution as _Distribution

# globals
VERBOSE = False
CONFIG_FILE = 'config.txt'
WSDL_SEC = 'WSDL'
TYPES_SEC = 'TYPES'
PERSISTENT_SEC = 'Persistent'
GENERATED = 'pyGridWare.generated'


if sys.version_info < (2, 4):
    raise SystemExit("Python 2.4 or later is required")


def _pkg2file(name, asdir=False):
    """utility function, 
    name --  package/module name
    """
    if asdir: return join(*name.split('.'))
    return '%s.py' %join(*name.split('.'))


def _done():
    """utility function, has code been generated.
    """
    return isfile(_pkg2file('%s.__init__' %GENERATED))


def _create_generated_files(items, services=True, stubs=True, cp=None, web2=None):
    """utility function, regenerate modules with new genrator
    """
    if VERBOSE:
         from ZSI.wstools.logging import setBasicLoggerDEBUG
         setBasicLoggerDEBUG()
         # need to initialize distutils log for setuptools
         log.set_verbosity(log.DEBUG)

    from ZSI.wstools.WSDLTools import WSDLReader
    from ZSI.wstools.XMLSchema import SchemaReader
    # set up pyclass metaclass
    from ZSI.generate.containers import TypecodeContainerBase, TypesHeaderContainer
    TypecodeContainerBase.metaclass = 'pyclass_type'
    TypesHeaderContainer.imports.append(\
        'from ZSI.generate.pyclass import pyclass_type'
    )
    sys.path.append(join(os.getcwd(), "pyGridWare", "utility", "generate"))
    from Modules import WriteServiceModules, WriteClientStubModule,\
        WriteTypesModule, NR
    from Utility import CLIENT,SERVICE,SERVICE_STUBS,TYPES,PROPERTIES,RESOURCE

    add_packages = NR.Reset()
    add_packages.insert(0, GENERATED)
    for opt,url in items:
        log.info("\t*** %-25s: %-50s",opt,url)
        wsdl = None
        reader = WSDLReader()
        if services is stubs is False:
            reader = SchemaReader()

        if isfile(url) is False:
            wsdl = reader.loadFromURL(url)
        else:
            wsdl = reader.loadFromFile(url)

        if services is stubs is False:
            schema = wsdl
            class wsdl:
                name = opt
                types = {schema.targetNamespace:schema}

        kw = {'web2':web2}
        if cp is not None:
            try: kw['simple'] = not cp.getboolean(PERSISTENT_SEC, opt)
            except NoOptionError: pass
            
        if services:
            WriteServiceModules(opt, wsdl, **kw)
        if stubs:
            WriteClientStubModule(opt, wsdl)
        else:
            WriteTypesModule(opt, wsdl)

    return add_packages
    
   
def _make_generated_package():
    """Returns False if "generated" package is already present,
    else generates it and returns True. 
    """
    if not (not _done() and len(sys.argv) > 1 and 
        sys.argv[1] in ('install', 'build', 'doc')):
        return False

    cp = ConfigParser()
    cp.read(CONFIG_FILE)

    # attempt to remove generated
    d = _pkg2file(GENERATED)
    if isdir(d):
        try:
            _remove_generated_files(d)
        except OSError:
            print "ERROR: Must Manually Remove Generation Directory (rm -rf %s)" %d
            sys.exit(-1)
 
    # Create types
    log.info("|" + "--"*22 + "|")
    log.info("|  %-40s  |" %"<< Core XSD Schema code generation >>")
    log.info("|" + "--"*22 + "|")
    _create_generated_files(cp.items(TYPES_SEC), services=False, stubs=False, cp=cp, web2=False)
        
    # Create services/stubs/types
    log.info("|" + "--"*22 + "|")
    log.info("|  %-40s  |" %"<< Core WSDL code generation >>")
    log.info("|" + "--"*22 + "|")
    _create_generated_files(cp.items(WSDL_SEC), services=True, stubs=True, cp=cp, web2=False)

    log.info("|" + "--"*30)
      
    return True 


def _remove_generated_files(d):
    """
    Utility method to clean up the geneated files.
    """
    for root, dirs, files in os.walk(d, topdown=False):
        for f in files:
            os.remove(os.path.join(root, f))
        for d in dirs:
            os.rmdir(os.path.join(root, d))

    os.rmdir(root)

 
class PGDistribution(_Distribution):
    """Indirection so that setuptools grabs "setup_requires" dependencies
    before calling "_make_generated_package()"
    """
    mine = True
    def __init__ (self, attrs=None):
        if self.__class__.mine is True:
            self.mine = True
            self.__class__.mine = False
 
        _Distribution.__init__ (self, attrs=attrs)

    def finalize_options (self):
        if self.mine is True and _make_generated_package() is True:
             self.packages = find_packages()
        _Distribution.finalize_options(self)


def main():
    # HACK Until > setuptools 0.6.3c
    from setuptools import sandbox
    sandbox.DirectorySandbox._ok = lambda self,path: True

    setup(name="pyGridWare", 
          version="1.2.0", 
          author="Joshua Boverhof", 
          author_email="JRBoverhof@lbl.gov",
          url="http://dsd.lbl.gov/gtg/projects/pyGridWare",
          download_url="http://dsd.lbl.gov/gtg/projects/download/download_info.html",
          description="Python WS-Resource Framework and GSS Security",
          maintainer="Joshua Boverhof", 
          maintainer_email="JRBoverhof@lbl.gov",

          classifiers = [
              "Development Status :: 5 - Production/Stable",
              "Intended Audience :: Developers",
              "Intended Audience :: Science/Research",
              "License :: OSI Approved :: BSD License",
              "Programming Language :: Python",
              "Natural Language :: English",
              "Operating System :: Microsoft :: Windows",
              "Operating System :: POSIX"
          ],
          packages=find_packages(),

          data_files=[
              ("config",['bin/server-config.tac', 'bin/server-config.txt',]),
          ],
          scripts=["tools/wsdl2web",
                   ],

          zip_safe=False,
          distclass=PGDistribution,

          entry_points = {\
          'distutils.commands': ['pydoc = pyGridWare.utility.command.docker:PyDocForGeneratedModules'],
          'console_scripts': ['pyrft = pyGridWare.clients.pyrft:main',
                              'pydelegate_credential = pyGridWare.clients.pydelegate_credential:main',
                              'pywsrf_query = pyGridWare.clients.pywsrf_query:main']
          },
          
          install_requires=[
              "ZSI >= 2.1.dev-r1352", "4Suite-XML >= 1.0rc3"
          ],

          extras_require={ 
              "MessageSecurity":["M2Crypto>=0.13", "pyGsi>=0.2"],
              "https":["M2Crypto>=0.13",], 
          },

          setup_requires=[
              #"ZSI >= 1.2.0rc2", "Twisted >= 2.0", "4Suite-XML >= 1.0rc3", 
              "ZSI >= 2.1.dev-r1352",
          ],

          dependency_links = [
              "http://sourceforge.net/project/showfiles.php?group_id=26590&package_id=30660",
              #"ftp://ftp.4suite.org/pub/4Suite/",
          ],
    )


if __name__ == '__main__': 
    main()
